import datetime
import sys
from .formatting import format_time
from .schema import wrap_exception, wrap_result


class Manager:
    """
    Wrap a bit of cloud or batch computation, automatically handling and
    serializing errors and timing the computation.

    Always returns a dictionary with the following keys:

    - `success` (`bool`): `True` so long as no exception was raised.
    - `result` (`object`): In case of succes, the deserialized result.
    - `error` (`None` or `list`): In case of an error, the formatted exception
        traceback.
    - `error_origin` (`None` or `str`): In case of an error, the source of the
        error, which is `'compute'` for errors which originate inside the
        context handler. Other values are reserved (but unsued by this
        function): `'system'` for errors on the compute system which are
        external to the compute code, and `'orchestration'` for other errors
        which occur while attempting to reach the remote system, including
        timeouts.
    - `start_time` (`str`): The local start time of the computation, in ISO
        8601 with the time zone.
    - `duration_seconds` (`float`): The time spent in computation.
    - `runtime_info` (`object`): User-provided runtime metadata.

    Args:
        handle_exceptions (bool): When `True`, exceptions thrown within the
            block are caught and serialized, as is useful for wrapping cloud
            processes or batch jobs, where you don't want to crash. When
            `False` these exceptions propagate, as is useful for development.
            `KeyboardInterrupt` and `SystemExit` exceptions always propagate.
        vebose (bool): When `True`, print out timing information when
            computation finishes.
        time_precision (int): The number of decimal places (of seconds) to
            include in the timing result. The default is 2, which rounds to
            the nearest hundredth of a second.

    Example:

        with Manager(handle_exceptions=handle_exceptions) as manager:
            result = perform_computation()
            # If needed, convert result to JSON-serializable objects.
            as_native = transform(result)
            manager.result = as_native
        return manager.serialized_result
    """

    def __init__(
        self, runtime_info=None, handle_exceptions=True, verbose=False, time_precision=2
    ):
        self.runtime_info = runtime_info
        self.handle_exceptions = handle_exceptions
        self.verbose = verbose
        self.time_precision = time_precision
        self.result = None

    def __enter__(self):
        self.start_time = datetime.datetime.now()
        return self

    def __exit__(self, type, value, traceback):
        duration_seconds = round(
            (datetime.datetime.now() - self.start_time).total_seconds(),
            self.time_precision,
        )

        if type in [KeyboardInterrupt, SystemExit]:
            raise value

        if value:
            self.serialized_result = wrap_exception(
                exception=value,
                error_origin="compute",
                start_time=self.start_time,
                duration_seconds=duration_seconds,
                runtime_info=self.runtime_info,
            )
            if self.handle_exceptions:
                print(
                    "Error handled by werkit. (To disable, invoke `Manager()` with `handle_exceptions=False`.)"
                )
                print("".join(self.serialized_result["error"]))
                return True
            else:
                return False
        else:
            self.serialized_result = wrap_result(
                serializable_result=self.result,
                start_time=self.start_time,
                duration_seconds=duration_seconds,
                runtime_info=self.runtime_info,
            )

        if self.verbose:
            print(
                "Completed in {}".format(format_time(duration_seconds)), file=sys.stderr
            )



from .cloud_manager import CloudManager, Config  # noqa: F401
from .invoke import (  # noqa: F401
    DEFAULT_QUEUE_NAME,
    clean,
    get_aggregate_status,
    get_all_jobs,
    get_results,
    invoke_for_each,
)



import asyncio
import datetime
import json
from functools import partial
import boto3
from botocore.exceptions import ClientError
from harrison import Timer


lambda_client = boto3.client("lambda")
event_loop = asyncio.get_event_loop()


async def call_worker_service(
    lambda_worker_function_name,
    extra_args,
    _input,
    with_timing=True,
    event_loop=event_loop,
    executor=None,
):
    invocation_start_timestamp = datetime.datetime.utcnow().timestamp()
    invoke = partial(
        lambda_client.invoke,
        FunctionName=lambda_worker_function_name,
        Payload=json.dumps({"input": _input, "extra_args": extra_args}),
    )
    with Timer(verbose=False) as response_timer:
        response = await event_loop.run_in_executor(executor, invoke)
        payload = await event_loop.run_in_executor(executor, response["Payload"].read)
    result = json.loads(payload.decode())
    if with_timing:
        result["invocation_start_timestamp"] = invocation_start_timestamp
        result["lambda_roundtrip_seconds"] = response_timer.elapsed_time_s
    return result


async def wait_for(
    timeout,
    lambda_worker_function_name,
    extra_args,
    _input,
    with_timing=True,
    event_loop=event_loop,
    executor=None,
):
    try:
        return await asyncio.wait_for(
            call_worker_service(
                lambda_worker_function_name,
                extra_args,
                _input,
                event_loop=event_loop,
                executor=executor,
                with_timing=with_timing,
            ),
            timeout=timeout,
        )
    except ClientError as ex:
        return ex
    except asyncio.TimeoutError as ex:
        return ex
    except Exception as ex:
        return ex


async def parallel_map_on_lambda(
    lambda_worker_function_name,
    timeout,
    input,
    extra_args=[],
    with_timing=True,
    event_loop=event_loop,
    executor=None,
):
    coroutines = [
        wait_for(
            timeout=timeout,
            lambda_worker_function_name=lambda_worker_function_name,
            _input=item,
            extra_args=extra_args,
            with_timing=with_timing,
            event_loop=event_loop,
            executor=executor,
        )
        for item in input
    ]
    return await asyncio.gather(*coroutines, return_exceptions=True)



from setuptools import find_packages, setup


def load(filename):
    return open(filename, "rb").read().decode("utf-8")


version_info = {}
exec(open("werkit/package_version.py").read(), version_info)


setup(
    name="werkit",
    version=version_info["__version__"],
    description="Toolkit for encapsulating Python-based computation into deployable and distributable tasks",
    long_description=load("README.md"),
    long_description_content_type="text/markdown",
    author="Metabolize",
    author_email="github@paulmelnikow.com",
    url="https://github.com/metabolize/werkit",
    project_urls={
        "Issue Tracker": "https://github.com/metabolize/werkit/issues",
        "Documentation": "https://werkit.readthedocs.io/en/stable/",
    },
    packages=find_packages(),
    install_requires=load("requirements.txt"),
    extras_require={
        "client": load("requirements_client.txt"),
        "aws_lambda_build": load("requirements_aws_lambda_build.txt"),
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: System :: Distributed Computing",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
)
